-- Beycan KAHRAMAN - 040020337 - 29.04.2007
-- diff - ADVANCED PROGRAMMING - Haskell 

import Array  -- used for character name table.
import Numeric (readDec, readOct, lexDigits, readHex)
import Char
import Text.Regex
import Maybe( listToMaybe )
import System( getArgs )

type FileDiff = [String]

printDiff :: FileDiff -> String -> Int -> IO () -- print list (look for the parameter to print)
printDiff differ s i = 
	if(i > 0) then 
		(putStr . concat . map (++"\n")) (["======= " ++ s ++ " ======================================"] ++ differ ++ [""])
	else
		putStr "";

lower :: FileDiff -> FileDiff -- case-sensitivity
lower differ = (map (map toLower)) differ

white :: FileDiff -> FileDiff -- remove white spaces
white differ = (map (unwords . words) ) differ

isIn :: String -> FileDiff -> Bool -- is the first list inside the second list ?
isIn a [] = False
isIn a b = 
	if (a == (head b)) then True
	else isIn a (tail b)

-- UNIFIED list creator ---
uContext :: FileDiff -> FileDiff -> FileDiff -> FileDiff -> FileDiff
uContext [] [] a b = []
uContext [] differ a b = (map ("+\t" ++)) b
uContext differ [] a b = (map ("-\t" ++)) a
uContext differ1 differ2 a b = 
	if (head differ1 == head differ2) then 
		["\t" ++ head a] ++ uContext (tail differ1) (tail differ2) (tail a) (tail b)
	else if ( isIn (head differ1) (tail differ2) ) then
		["+\t" ++ head b] ++ uContext (differ1) (tail differ2) (a) (tail b)
	else
		["-\t" ++ head a] ++ uContext (tail differ1) (differ2) (tail a) (b)

-- CONTEXT list creator ---
cContext :: [Char] -> FileDiff -> FileDiff -> FileDiff -> FileDiff -> FileDiff
cContext c [] [] a b = []
cContext c [] differ a b = (map ((c ++ "\t") ++)) b
cContext c differ [] a b = []
cContext c differ1 differ2 a b = 
	if (head differ1 == head differ2) then 
		["\t" ++ head a] ++ cContext c (tail differ1) (tail differ2) (tail a) (tail b)
	else if ( isIn (head differ1) (tail differ2) ) then
		[c ++ "\t" ++ head b] ++ cContext c (differ1) (tail differ2) (a) (tail b)
	else
		cContext c (tail differ1) (differ2) (tail a) (b)

-- NORMAL list creator ---
nContext :: FileDiff -> FileDiff -> FileDiff -> FileDiff -> FileDiff
nContext [] [] a b = []
nContext [] differ a b = (map ("+\t" ++)) b
nContext differ [] a b = (map ("-\t" ++)) a
nContext differ1 differ2 a b = 
	if (head differ1 == head differ2) then 
		--["\n"] ++ nContext (tail differ1) (tail differ2) (tail a) (tail b)
		nContext (tail differ1) (tail differ2) (tail a) (tail b)
	else if ( isIn (head differ1) (tail differ2) ) then
		["+\t" ++ head b] ++ nContext (differ1) (tail differ2) (a) (tail b)
	else
		["-\t" ++ head a] ++ nContext (tail differ1) (differ2) (tail a) (b)

-- CALCULATE parameters ---
parameters :: [String] -> String -> Int
parameters [] s = 0
parameters arg s = 
	if ( head arg == s ) then
		1 + parameters (tail arg) s
	else
		parameters (tail arg) s

-- if parameters given, change the list
changeWhiteList :: Bool -> [String] -> [String]
changeWhiteList a b = if (a) then white b else b

changeCaseList :: Bool -> [String] -> [String]
changeCaseList a b = if (a) then lower b else b
		

-- MAIN FUNCTION ---
main = do
	args <- getArgs
	
	case (length args < 2) of --(I COULD NOT EXIT according to args)
		True -> print "USAGE file1 file2 [parameters]"
		False -> putStrLn ""
	
	param3 <- return (parameters args "-u")
	param4 <- return (parameters args "-c")
	
	case (param3 + param4 == 2) of --(I COULD NOT EXIT according to args)
		True -> print "confusing parameters (-c & -u)"
		False -> putStr ""
	
	param1 <- return (parameters args "-i")
	param2 <- return (parameters args "-w")
	
	file1 <- return (head args)
	file2 <- return (head(tail args))

	differ1 <- readFile file1
	differ2 <- readFile file2
	diffOriginalLines1 <- return (lines differ1)
	diffOriginalLines2 <- return (lines differ2)
    
	diffLines1 <- return diffOriginalLines1
	diffLines2 <- return diffOriginalLines2
	
	-- assuming that we have -i parameter
	diffLines1 <- return (changeCaseList (odd param1) diffOriginalLines1)
	diffLines2 <- return (changeCaseList (odd param1) diffOriginalLines2)
    	
    -- assuming that we have -w parameter 
	diffLines1 <- return (changeWhiteList (odd param2) diffLines1)
	diffLines2 <- return (changeWhiteList (odd param2) diffLines2)
    
    -- we have calculated all of the diff commands according to parameters
	resultN <- return ( nContext diffLines1 diffLines2 diffOriginalLines1 diffOriginalLines2)
	printDiff resultN "NORMAL" (1-param3 - param4) -- N
	
	resultU <- return ( uContext diffLines1 diffLines2 diffOriginalLines1 diffOriginalLines2)
	printDiff resultU "UNIFIED" param3 -- U
	
	resultC1 <- return ( cContext "-" diffLines1 diffLines2 diffOriginalLines1 diffOriginalLines2)
	printDiff resultC1 "CONTEXT" param4 -- C
	resultC2 <- return ( cContext "+" diffLines2 diffLines1 diffOriginalLines2 diffOriginalLines1)
	printDiff resultC2 "=======" param4
	