// Beycan KAHRAMAN - 040020337 - 29.04.2007
// diff - ADVANCED PROGRAMMING - C 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct node * initnode( char *, char *, int );
void printlist( struct node ** );
void add( struct node **, struct node * );
int readFile(struct node **, char *, char *);
void readParameters(int, char **, char *);
void deleteList( struct node ** );
int diff(struct node *, struct node *);
void printDiff(struct node *, struct node *, char *);

struct node { // data structure to handle differences & lines
    char *name;
    char *original; 
    short line_no;
    short error_no;
    short error_end_line;
    char error_type;
    char error_type_n;
    char error_type_c;
    char error_type_u;
    struct node *next;
};

// create new node -> will be used in list
struct node * initnode( char *name, char *original, int line_no )
{
    struct node *ptr;
    ptr = (struct node *) calloc( 1, sizeof(struct node ) );
    ptr->name = malloc(strlen(name) + 1);
    ptr->original = malloc(strlen(name) + 1);
    if( ptr == NULL )
        return (struct node *) NULL;
    else {
        strcpy( ptr->name, name );
        strcpy( ptr->original, original );
        ptr->line_no = line_no;
        ptr->error_no = 0;
        ptr->error_end_line = 0;
        ptr->error_type = ' ';
        ptr->error_type_n = ' ';
        ptr->error_type_c = ' ';
        ptr->error_type_u = ' ';
        return ptr;        
    }
}

// print list
void printlist( struct node **head )
{
    struct node *ptr = *head; 

    while( ptr != NULL )
    {
        printf("%d %c %d %s", ptr->line_no, ptr->error_type, ptr->error_no, ptr->original );
        ptr = ptr->next; 
    }
    printf(" >> [end of file]\n\n");
}

// add node to list
void add( struct node **head, struct node *new ) 
{
    struct node *tmp;
    if( *head == NULL )
         *head = new;   
    else{
        tmp = *head;
        while(tmp->next)
            tmp = tmp->next;
        tmp->next = new;
    }    
}

// white space remover
char * removeWhiteSpaces(char *c){
    int i, j = 0, len = strlen(c);
    char *d = malloc(len + 1);
    
    for(i=0; i<len; ++i){
        if(c[i] == ' ' && c[i+1] == ' ')
            continue;
        d[j++] = c[i];
    }  
    d[j] = '\0';

    return d;      
}

// used to disable case sensitivity
char * toUpper(char *c){
    int i, len = strlen(c);
    char *d = malloc(len + 1);
    
    for(i=0; i<len; ++i){
        if(c[i] >= 'a' && c[i] <= 'z')
            d[i] = c[i] + 'A' - 'a';
        else
            d[i] = c[i];
    }  
    d[i] = '\0';

    return d;  
}

// reads the given file to initial list
int readFile(struct node **file_list, char *file_name, char *parameters){
    FILE *f;
    int i = 0;
    char str[1024], tmp[1024];
    struct node *ptr;

    f = fopen(file_name, "r");
    if(!f)
        return 0;

    while(!feof(f)){
        strcpy(tmp, str);
        fgets(str, 1024, f);
        if(feof(f) && strcmp(tmp, str) == 0)
            break;
        ++i;
        if(parameters[0] == '+' && parameters[1] == '+')
            ptr = initnode( removeWhiteSpaces(toUpper(str)), str, i );
        else if(parameters[0] == '+')
            ptr = initnode( toUpper(str), str, i );
        else if(parameters[1] == '+')
            ptr = initnode( removeWhiteSpaces(str), str, i );
        else
            ptr = initnode( str, str, i );
        add( file_list, ptr );
    }
    //printlist( file_list );
    return 1;
}

// look for parameters
void readParameters(int argc, char **argv, char *parameters){
    int i;
    for(i=3; i<argc; ++i)
        if(strcmp(argv[i], "-i") == 0)
            parameters[0] = '+';
        else if(strcmp(argv[i], "-w") == 0)
            parameters[1] = '+';
        else if(strcmp(argv[i], "-u") == 0)
            parameters[2] = '+';
        else if(strcmp(argv[i], "-c") == 0)
            parameters[3] = '+';
    //printf("%s\n", parameters);
}

// delete whole list
void deleteList( struct node **head ) 
{
    struct node *del, *tmp;

    if( *head != NULL ){
        del = *head;
        tmp = del;

        while(del = tmp){
            tmp = del->next;
            if(del)
                free(del);
        }
    }    
}

// calculate diff between two files
int diff(struct node *file_list_1, struct node *file_list_2){
    struct node *errorStart1 = file_list_1, *errorStart2 = file_list_2;
    struct node *errorEnd1 = file_list_1, *errorEnd2 = file_list_2;
    int error_no = 0;
    int error_end_line1, error_end_line2;
    char error_type;
    
    while(errorStart1 || errorStart2){
        if( !errorStart1 || !errorStart2 || strcmp(errorStart1->name, errorStart2->name) != 0){
            errorEnd1 = errorStart1;
            errorEnd2 = errorStart2;
            while( 1 ){
                
                if(!errorEnd1 || !errorEnd2 || strcmp(errorEnd1->name, errorEnd2->name) == 0){
                    if(!errorEnd1) errorEnd2 = NULL;
                    if(!errorEnd2) errorEnd1 = NULL;
                    ++error_no;

                    if(!errorEnd1) {
                        error_end_line1 = -1;
                        error_end_line2 = -1; 
                    }else {
                        error_end_line1 = errorEnd1->line_no;
                        error_end_line2 = errorEnd2->line_no;
                    }

                    if(errorStart1 == errorEnd1) error_type = 'a';
                    else if(errorStart2 == errorEnd2) error_type = 'd';
                    else error_type = 'c';

                    if(errorStart1) errorStart1->error_no = error_no;

                    while(errorStart1 != errorEnd1 && errorStart1){
                        errorStart1->error_no = error_no;
                        errorStart1->error_type = error_type;
                        switch(error_type){
                            case 'a':   errorStart1->error_type_n = '>'; 
                                        errorStart1->error_type_c = '-'; 
                                        errorStart1->error_type_u = '+'; break;
                            case 'd':   errorStart1->error_type_n = '<'; 
                                        errorStart1->error_type_c = '-'; 
                                        errorStart1->error_type_u = '-'; break;
                            case 'c':   errorStart1->error_type_n = '<'; 
                                        errorStart1->error_type_c = '!'; 
                                        errorStart1->error_type_u = '-'; break;
                        }
                        errorStart1->error_end_line = error_end_line1;
                        errorStart1 = errorStart1->next;
                    }

                    if(errorStart2) errorStart2->error_no = error_no;

                    while(errorStart2 != errorEnd2 && errorStart2){
                        errorStart2->error_no = error_no;
                        errorStart2->error_type = error_type;
                        switch(error_type){
                            case 'a':   errorStart2->error_type_n = '>'; 
                                        errorStart2->error_type_c = '+'; 
                                        errorStart2->error_type_u = '+'; break;
                            case 'd':   errorStart2->error_type_n = '<'; 
                                        errorStart2->error_type_c = '-'; 
                                        errorStart2->error_type_u = '-'; break;
                            case 'c':   errorStart2->error_type_n = '>'; 
                                        errorStart2->error_type_c = '!'; 
                                        errorStart2->error_type_u = '+'; break;
                        }
                        errorStart2->error_end_line = error_end_line2;
                        errorStart2 = errorStart2->next;
                    }
                    break;
                }

                errorEnd1 = errorEnd1->next;
                if(!errorEnd1){
                    errorEnd2 = errorEnd2->next;
                    errorEnd1 = errorStart1;
                }
            }
        }
        if( errorStart1 ) errorStart1 = errorStart1->next;
        if( errorStart2 ) errorStart2 = errorStart2->next;
    }
    
    return 1;
}

// print calculated diff
void printDiff(struct node *file_list_1, struct node *file_list_2, char *parameters){
    struct node *ptr1 = file_list_1, *ptr2 = file_list_2, *tmp = NULL;
    int last_error_no;
    char last_error_type;
    int write_back;
    printf("\n");

    if(parameters[3] == '+'){ // -c
        while(ptr1){
            if(ptr1->error_no > 0){
                printf("\t '  '  '  '  '  '\n");
                if(tmp)
                    printf("%d %c\t %s", tmp->line_no, tmp->error_type_c, tmp->original);
                if(ptr1 && (ptr1->original)[strlen(ptr1->original)-1] != '\n') printf(" [no endline]\n");
                while(ptr1 && ptr1->error_no > 0){
                    printf("%d %c\t %s", ptr1->line_no, ptr1->error_type_c, ptr1->original);
                    if(ptr1 && (ptr1->original)[strlen(ptr1->original)-1] != '\n') printf(" [no endline]\n");
                    ptr1 = ptr1->next;
                }
                if(ptr1 && ptr1->error_no < 1)
                    printf("%d %c\t %s", ptr1->line_no, ptr1->error_type_c, ptr1->original);
                if(ptr1 && (ptr1->original)[strlen(ptr1->original)-1] != '\n') printf(" [no endline]\n");
            }else{
                tmp = ptr1;
                ptr1 = ptr1->next;
            }
        }
        printf("-----------------------------------------------------------------\n");
        tmp = NULL;
        while(ptr2){
            if(ptr2->error_no > 0){
                printf("\t '  '  '  '  '  '\n");
                if(tmp)
                    printf("%d %c\t %s", tmp->line_no, tmp->error_type_c, tmp->original);
                if(ptr2 && (ptr2->original)[strlen(ptr2->original)-1] != '\n') printf(" [no endline]\n");
                while(ptr2 && ptr2->error_no > 0){
                    printf("%d %c\t %s", ptr2->line_no, ptr2->error_type_c, ptr2->original);
                    if(ptr2 && (ptr2->original)[strlen(ptr2->original)-1] != '\n') printf(" [no endline]\n");
                    ptr2 = ptr2->next;
                }
                if(ptr2 && ptr2->error_no < 1)
                    printf("%d %c\t %s", ptr2->line_no, ptr2->error_type_c, ptr2->original);
                if(ptr2 && (ptr2->original)[strlen(ptr2->original)-1] != '\n') printf(" [no endline]\n");
            }else{
                tmp = ptr2;
                ptr2 = ptr2->next;
            }
        }
    }else{ // normal & unified (in normal do not print non-necessary lines)
        while(ptr1){
            if(ptr1->error_no > 0){
                last_error_no = ptr1->error_no;
                last_error_type = ptr1->error_type;
                
                write_back = 0;
                tmp = ptr1;
                if(tmp && (tmp->error_type == 'd' || tmp->error_type == 'c'))
                    while(tmp && tmp->error_no == last_error_no){
                        if(parameters[2] == '+')
                            printf("%d %c\t %s", tmp->line_no, tmp->error_type_u, tmp->original);
                        else
                            printf("f1 %d %c\t %s", tmp->line_no, tmp->error_type_n, tmp->original);
                        if(tmp && (tmp->original)[strlen(tmp->original)-1] != '\n') printf(" [no endline]\n");
                        tmp = tmp->next;
                    }
                ptr1 = tmp;
                if(ptr1 && ptr1->error_no == last_error_no)
                    write_back = 1;
                if(parameters[2] != '+')
                    printf("-----------------------------------------------------------------\n");
                tmp = ptr2;
                while(tmp && tmp->error_no != last_error_no)
                    tmp = tmp->next;
                if(tmp && (tmp->error_type == 'a' || tmp->error_type == 'c')){
                    while(tmp && tmp->error_no == last_error_no){
                        if(parameters[2] == '+')
                            printf("%d %c\t %s", tmp->line_no, tmp->error_type_u, tmp->original);
                        else
                            printf("f2 %d %c\t %s", tmp->line_no, tmp->error_type_n, tmp->original);
                        if(tmp && (tmp->original)[strlen(tmp->original)-1] != '\n') printf(" [no endline]\n");
                        tmp = tmp->next;
                    }
                }
                ptr2 = tmp;
                if(write_back){
                    if(parameters[2] == '+')
                        printf("%d %c\t %s", ptr1->line_no, ptr1->error_type_u, ptr1->original);
                    ptr1 = ptr1->next;
                }
                if(parameters[2] != '+')
                    printf("\n");
            }else{
                if(parameters[2] == '+')
                    printf("%d %c\t %s", ptr1->line_no, ptr1->error_type_u, ptr1->original);
                ptr1 = ptr1->next;
            }
        }
        if(!ptr1){
            if(ptr2 && ptr2->error_no > 0) ptr2 = ptr2->next;
            tmp = ptr2;
            if(parameters[2] != '+')
                while(tmp && tmp->error_no < 1)
                    tmp = tmp->next;
            if(tmp){
                if(parameters[2] != '+')
                    printf("-----------------------------------------------------------------\n");
                while(tmp){
                    if(parameters[2] == '+' && tmp->error_no > 0)
                        printf("%d %c\t %s", tmp->line_no, tmp->error_type_u, tmp->original);
                    else if(tmp->error_no > 0)
                        printf("f2 %d %c\t %s", tmp->line_no, tmp->error_type_n, tmp->original);
                    if(tmp && (tmp->original)[strlen(tmp->original)-1] != '\n' && tmp->error_no > 0) printf(" [no endline]\n");
                    tmp = tmp->next;
                }
            }
        }
    }
    if(parameters[2] == '+')
        printf("\n");
}

// MAIN FUNCTION
int main(int argc, char *argv[])
{
	char c[100];
	strcpy(c, " df dsg sdf       df sdfh      sdh dgh gh           gf");
	printf("%s\n%s\n", c, removeWhiteSpaces(c));

/*
    int total_changes = 0;
    struct node *file_list_1 = (struct node *) NULL;
    struct node *file_list_2 = (struct node *) NULL;
    char parameters[5];               // -i -w -u -c
    strcpy(parameters, "----\0");      // all false

    if(argc < 3){
        printf("Usage: Program File1 File2 [parameters]\n\n");
        return -1;
    }

    if(argc > 3)
        readParameters(argc, argv, parameters);

    if(parameters[2] == '+' && parameters[3] == '+'){
        printf("Confusing Parameters ( -c & -u ) ...\n\n");
        return -1;
    }

    if(readFile(&file_list_1, argv[1], parameters) == 0 || readFile(&file_list_2, argv[2], parameters) == 0){
        printf("Files Could Not be Found...\n\n");
        return -1;
    }

    total_changes = diff(file_list_1, file_list_2); // find total changes in files

    printDiff(file_list_1, file_list_2, parameters); // print diff

    deleteList(&file_list_1);
    deleteList(&file_list_2);

    return total_changes;*/
}
